#ifndef DYNAMICUPDATECONTROL_H
#define DYNAMICUPDATECONTROL_H

#include <map>
#include <string>
#include <sstream>
#include <iostream>

/**
 * \class DynamicUpdateControl
 * The DynamicUpdateControl class directly updates parameters according to strings as they are received by Tekkotsu's TextMsgEvent; any string in the proper format can be used, of course. Be careful that any variables registered to a DynamicUpdateControl object are alive at least as long as it is. Also, this class will not release any memory, besides what it allocates internally.
 * Herein, the term "variable" is the entity external to DynamicUpdateControl, and "parameter" is what DynamicUpdateControl has a hold of.  When registering a variable, it's not necessary to give its name as the parameter name. For example, you can register "distanceControllerDeltaError" as "deltaE" or "e", so long as no other parameters have registered those names.
 * I believe that this class will support any type that has implemented operator<<, operator>>, operator=, and a copy constructor, though I haven't tested it because I was concerned with getting native typed parameters (int, float, etc.) to work.
 * \author Luis Guimbarda
 * \date 9/25/2009
 */
class DynamicUpdateControl {
public:
	/**
	 * The constructor for DynamicUpdateControl. Takes an ostream reference for output (default is cout).
	 */
	DynamicUpdateControl(std::ostream& outputStream = std::cout);

	/**
	 * The destructor for DynamicUpdateControl. Releases all memory (except for pointers to external locations). Does NOT close the output ostream.
	 */
	~DynamicUpdateControl();

	/**
	 * This saves a mapping from prmName to prmPtr. Parameter names can't include whitespace.
	 *
	 * \param	prmName	The name of the parameter
	 * \param	prmPtr	A pointer to the variable that is to become a parameter
	 *
	 * \exception	ParameterAlreadyRegistered	If another parameter with the same name was previously registered
	 */
	template<class T>
	void registerParameter(const std::string& prmName, T* prmPtr);

	/**
	 * Takes a string command, parses it, and processes it accordingly.
	 *	 The following is a list of actions and their command strings. Tekkotsu's ControllerGui textbox requires you prepend "!msg " to any string carried by its TextMsgEvent, but DynamicUpdateControl only expects the command strings listed below (which is, fortunately, what TextMsgEvent delivers). In all cases, <prmName> and <prmValue> may be omitted. Where they are, <prmName> is taken to be the currentWorkingParameter (i.e., the last parameter for which any action was taken), and <prmValue> is taken as the current value for <prmName>.
	 *	 - set/show parameter:			<prmName> <prmValue>
	 *	 - open parameter file:			#open <filename>
	 *	 - save parameter file:			#save <filename>
	 *	 - set parameter:				#set <prmName> <prmValue>
	 *	 - show parameter's value:		#show <prmName>
	 *	 - show all parameters' values:	#showAll
	 *
	 * \param	msg	The string (delivered by Tekkotsu's TextMsgEvent) containing a DynamicUpdateControl command
	 *
	 * \exception	MalformedMessage		If msg is not in a recognizable form
	 * \exception	ParseError				If prmValue cannot be parse as the expected type, or a value can be parsed but not with exausting the command string
	 * \exception	UnregisteredParameter	If prmName is not a registered parameter name (see registerParameter)
	 * \exception	others...
	 */
	void processMessage(const std::string& msg);

	class ParameterAlreadyRegistered {
	};
	class MalformedMessage {
	};
	class UnregisteredParameter {
	};
	class ParseError {
	};
	class UnrecognizedCommand {
	};

private:
	/**
	 *	 Updates the current value of prmName to the parsed prmValue.
	 *
	 *	 \param	prmName		The name of the parameter to be updated
	 *	 \param	prmValue	The string containing the parameter's new value
	 *
	 *	 \exception	ParseError				If prmValue cannot be parsed as its expected type
	 *	 \exception	UnregisteredParameter	If prmName is not a registered parameter name (see registerParameter)
	 */
	void updateParameter(const std::string& prmName,
			const std::string& prmValue);

	/**
	 *	 Sets the current and default values of all parameters as defined in the given filename. Only parameter updates will be processed, everything else is skipped. Any parameters listed in the file that aren't registered will be skipped. Any parameters whose value and/default can't be parsed will be skipped. A message will be output for any skipped lines. Multiple parameter descriptions are OK (since saveParameterFile appends), but only the last description is used. (Technically, all are used and overwritten but the last, which shouldn't be a problem barring weird side effects.)
	 *	 Each line is in the following format: "prmName prmValue prmDefault";
	 *	 Not yet implemented.
	 *
	 *	 \param	filename	Name of the file to be opened
	 *
	 *	 \exception	fileiostuff	i'm sure...
	 */
	void openParameterFile(const std::string& filename);

	/**
	 *	 Writes each parameters name, value, and default to the given file. If the file exists, the current parameters data are appended to the file.
	 *	 Each line is in the following format: "prmName prmValue prmDefault".
	 *	 Not yet implemented.
	 *
	 *	 \param	filename	Name of the file to which is written
	 *
	 *	 \exception	fileiostuff	i'm sure...
	 */
	void saveParameterFile(const std::string& filename);

	/**
	 *	 Outputs the parameter prmName's data as: "name value(default)".
	 *
	 *	 \exception	UnregisteredParameter	If prmName is not a registered parameter name (see registerParameter)
	 */
	void showParameter(const std::string& prmName);

	/**
	 *	 Outputs all parameters' data as in showParameter.
	 */
	void showAllParameters();

	/**
	 *	 \class BaseHandle
	 *	 An abstract class from which all ParameterHandle instantiations are derived.
	 */
	class BaseHandle {
	public:
		virtual void read(const std::string& s) = 0;
		virtual std::string toString() const = 0;
	};

	/**
	 *	 \class ParameterHandle
	 *	 A templated handle for parameters in DynamicUpdateControl objects.
	 */
	template<class T>
	class ParameterHandle: public BaseHandle {
	private:
		T* prmPtr;
	public:
		ParameterHandle(T* tPtr);

		void read(const std::string& s);
		std::string toString() const;
	};

	/**
	 *	 A map from parameter names to ParamaterHandles of pointers to their current values. The pointers are acquired via registerParameter.
	 */
	std::map<std::string, BaseHandle*> ptrTbl;

	std::string cwp; /**< The current working parameter. */
	std::ostream& out; /**< The ostream to which output is sent. Defaults to std::cout. */
};

template<class T>
void DynamicUpdateControl::registerParameter(const std::string& prmName,
		T* prmPtr) {
	if (ptrTbl.count(prmName) > 0) {
		throw ParameterAlreadyRegistered();
	}
	ptrTbl[prmName] = new ParameterHandle<T> (prmPtr);
	cwp = prmName;
}
template<class T>
DynamicUpdateControl::ParameterHandle<T>::ParameterHandle(T* tPtr) :
	prmPtr(tPtr) {
}

template<class T>
void DynamicUpdateControl::ParameterHandle<T>::read(const std::string& s) {
	std::istringstream iss(s);
	T t; // used to prevent partial reading updating parameters
	char ch;

	if ((iss >> t).fail() || !(iss >> ch).fail()) {
		// we can't parse the string or we can, but there is more string remaining
		throw DynamicUpdateControl::ParseError();
	}
	*prmPtr = t;
}

template<>
void DynamicUpdateControl::ParameterHandle<bool>::read(const std::string& s) {
	std::istringstream iss(s);
	bool t; // used to prevent partial reading updating parameters
	char ch;

	if ((iss >> std::boolalpha >> t).fail() || !(iss >> ch).fail()) {
		// we can't parse the string or we can, but there is more string remaining
		throw DynamicUpdateControl::ParseError();
	}
	*prmPtr = t;
}

template<>
void DynamicUpdateControl::ParameterHandle<std::string>::read(const std::string& s){
	*prmPtr = s;
}

template<class T>
std::string DynamicUpdateControl::ParameterHandle<T>::toString() const {
	std::ostringstream oss(std::ostringstream::out);
	oss << *prmPtr;
	return oss.str();
}
template<>
std::string DynamicUpdateControl::ParameterHandle<bool>::toString() const {
	std::ostringstream oss(std::ostringstream::out);
	oss << std::boolalpha << *prmPtr;
	return oss.str();
}

#endif
